/*
 * Network utilities
 */

#ifndef ___NET_H_
#define ___NET_H_

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff     /* -1 return */
#endif /* INADDR_NONE */


/* A open_server_socket() opens the socket on TCP/IP AF_INET SOCK_STREAM.
 * In the function, it calls socket(), bind(), and listen.
 * If it is successfully opend, it returns the socket descriptor,
 * otherwise returns negative value.
 */
int open_server_socket(unsigned short tcp_port);

/* A connect_to_server() connects to the server which is specified with
 * hostname and tcp_port. (The socket is opend with AF_INET/SOCK_STREAM)
 * If it is success, it returns the opened socket discriptor, otherwise
 *  returns negative value.
 */
int connect_to_server(const char* hostname, unsigned short tcp_port);

void so_reuseaddr(int sfd);

void pack_sockaddr_in(struct sockaddr_in* sock_addr,
		      unsigned short port,
		      unsigned int   iaddr);

unsigned int name2addr(const char* name);
int bind_in(int sfd, struct sockaddr_in* addr);
int connect_in(int fd, struct sockaddr_in* addr);
int recvfrom_in(int sfd, char* packet, int maxsize, struct sockaddr_in* from);
int sendto_in(int sfd, char* packet, int size, struct sockaddr_in* to);
int simple_send_udp(char* packet, int size,
		    unsigned short port, unsigned int ipaddr);
int accept_and_fork(int sfd, struct sockaddr_in* from);
struct hostent *local_hostent(void);
unsigned int my_ip_address(void);


/* read/write n bytes */
int nread(int fd, void* buff, unsigned long nbytes);
int nwrite(int fd, void* buff, unsigned long nbytes);

#endif /* ___NET_H_ */
