#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "net.h"
#include "tut.h"

#ifdef sun
extern int errno;
#endif /* sun */

#ifndef CHECK
#define CHECK 0
#endif

static int sfd;
char recv_packet[MAXPACKETSIZE+SPARESIZE];
char send_packet[MAXPACKETSIZE+SPARESIZE];
U16 tu_conn_port = DEFAULT_TU_CONN_PORT;

static void doit(int client_ifd, int client_ofd,
		 const char* hostname, U16 bind_port, U16 port, U32 tcp_addr);
static char *parse_relay_arg(const char* arg, U16* bind_port, U16* port);

void sigchld_handler(int sig)
{
    int status, pid;

    while((pid = wait3(&status, WNOHANG, NULL)) > 0)
    {
	printf("##############################\n");
	printf("## EXIT: %d\n", pid);
	printf("##############################\n");
    }
    signal(SIGCHLD, sigchld_handler);
}

void usage(void)
{
    fprintf(stderr, "Usage: tu [-h] [-relay <hostname:port>]\n");
    exit(1);
}

void main(int argc, char** argv)
{
    int i, fd;
    char buff[BUFSIZ], *p, *q, *hostname;
    U16 bind_port, port;
    U32 tcp_addr;
    int dflag = 0;

    hostname = NULL;
    bind_port = 0;

    /* parse args */
    for(i = 1; i < argc; i++)
    {
	if(!strcmp(argv[i], "-h"))
	{
	    usage();
	}
	else if(!strcmp(argv[i], "-relay"))
	{
	    if(i + 1 == argc)
		usage();
	    if((hostname = parse_relay_arg(argv[++i], &bind_port, &port))
	       == NULL)
		usage();
	}
	else if(!strcmp(argv[i], "-port"))
	{
	    if(i + 1 == argc)
		usage();
	    tu_conn_port = atoi(argv[++i]);
	}
	else if(!strcmp(argv[i], "-d"))
	{
	    dflag = 1;
	}
	else if(argv[i][0] == '-')
	{
	    usage();
	}
	else
	{
	    break;
	}
    }

    if(dflag)
	daemon_init();

    sfd = open_server_socket(tu_conn_port);
    if(sfd < 0)
    {
	perror("Can't make socket");
	exit(1);
    }

    signal(SIGCHLD, sigchld_handler);
    fd = accept_and_fork(sfd, NULL);
#if CHECK
    printf("TCP Connection is established\n");
#endif

    if(hostname != NULL)
    {
	doit(fd, fd, hostname, bind_port, port, 0);
	/* NOT REACHED */
    }

    /* hostname [[tcp_addr:]port[,bind_port]] */
    p = buff;
    port = 0;
    tcp_addr = 0;
    bind_port = 0;
    /* hostname ɤ߹ */
    while(1)
    {
	int c;

	c = fdgetc(fd);
	if(c == -1)
	    exit(1);

	if(c == '\n' || c == '\0')
	{
	    *p = '\0';
	    port = DEFAULT_TCP_PORT;
	    if(p == buff)
	    {
		strcpy(buff, "Usage: UT-hostname [[hostname:]port]\n");
		write(fd, buff, strlen(buff));
		close(fd);
		exit(1);
	    }
	    if(p[-1] == '\r')
		p[-1] = '\0';
	    break;
	}
	if(c == ' ')
	{
	    *p++ = '\0';
	    break;
	}
	*p++ = c;
    }

    hostname = buff;
    if(port == 0)
    {
	/* tcp_addr ɤ߹ */
	q = p;
	while(1)
	{
	    int c;

	    c = fdgetc(fd);
	    if(c == -1)
		exit(1);
	    if(c == '\n')
	    {
		port = atoi(q);
		tcp_addr = 0;
		break;
	    }
	    else if(c == ':')
	    {
		char* r;

		*p++ = '\0';

		tcp_addr = name2addr(q);
		if(tcp_addr == INADDR_NONE)
		{
		    shutdown(fd, 2);
		    exit(1);
		}

		/* port ɤ߹ */
		q = p;
		while((c = fdgetc(fd)) > 0)
		{
		    if(c == '\n')
			break;
		    *p++ = c;
		}
		*p++ = '\0';
		port = atoi(q);
		if((r = strchr(q, ',')) != NULL)
		    bind_port = atoi(r + 1);

		break;
	    }
	    *p++ = c;
	}

	if(port == 0)
	{
	    strcpy(buff, "Invalid port\n");
	    write(fd, buff, strlen(buff));
	    close(fd);
	    exit(1);
	}
    }

    doit(fd, fd, hostname, bind_port, port, tcp_addr);
    /* NOT REACHED */
}

static void doit(int rfd, int wfd,
		 const char* hostname, U16 bind_port, U16 port, U32 tcp_addr)
{
    enum tut_status_t status;
    int send_size, recv_size;
    unsigned long wait_usec_time;
    struct sockaddr_in maddr, yaddr, conn_addr;
    struct sockaddr_in recv_addr;
    U16 us, usdmy; /* 16 bit unsigned integer */
    U32 id;
    U32 yipaddr;
    U32 ipaddr;	/* ut ip address */
    U16 tut_port;	/* TUT local port */
    fd_set orig_fdset, fdset;	/* for select */
    int nfds, cnt;
    struct timeval timeout;
    int timeout_flag;
    int usfd;			/* UDP Socket file descripter */
    U16 mseq = 0;		/* ˼ ack      number */
    U16 yseq = 0;		/* ˼ sequence number */
    U16 recv_seq;		/* received seq number */
    U16 recv_ack;		/* received ack number */
    U32 recv_flags;		/* received flags */
    int resend_cnt;		/* β */
    int n;

#if CHECK
    printf("doit hostname:<%s>, port<%d>\n", hostname, port);
#endif

    if((ipaddr = name2addr(hostname)) == INADDR_NONE)
    {
	fdprintf(wfd, "Unknown host: <%s>\n", hostname);
	close(wfd); close(rfd); exit(1);
    }

#if 0 /* old version */
    /* MIN_UDP_PORT  MAX_UDP_PORT ޤǤΤ̤ѤΥݡȤ bind 롣*/
    for(tut_port = MIN_UDP_PORT; tut_port <= MAX_UDP_PORT; tut_port++)
    {
	printf("## check udp/%d...", tut_port); fflush(stdout);

	if((usfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
	    fdperror(wfd, "socket");
	    close(wfd); close(rfd); exit(1);
	}
	so_reuseaddr(usfd);
	pack_sockaddr_in(&maddr, tut_port, htonl(INADDR_ANY));
	if(bind(usfd, (struct sockaddr *)&maddr, sizeof(maddr)) >= 0)
	    break; /* success */

	if(errno != EADDRINUSE)
	{
	    fdperror(wfd, "bind");
	    close(usfd);
	    close(wfd); close(rfd); exit(1);
	}
	printf("## NG\n"); fflush(stdout);
	close(usfd);
    }
    if(tut_port > MAX_UDP_PORT)
    {
	fdprintf(wfd, "UDP Port is exhausted\n");
	close(wfd); close(rfd); exit(1);
    }
#else /* new version */
    if((usfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
	fdperror(wfd, "socket");
	close(wfd); close(rfd); exit(1);
    }    so_reuseaddr(usfd);
    pack_sockaddr_in(&maddr, 0, htonl(INADDR_ANY));

    if(bind(usfd, (struct sockaddr *)&maddr, sizeof(maddr)) < 0)
    {
	perror("bind");
	close(wfd); close(rfd); exit(1);
    }

    getsockname(usfd, (struct sockaddr *)&maddr, &n);
    tut_port = ntohs(maddr.sin_port);
#endif

    pack_sockaddr_in(&conn_addr, UU_CONN_PORT, ipaddr);

    /* fdset ν */
    FD_ZERO(&orig_fdset);
    FD_SET(usfd, &orig_fdset);
    nfds = usfd + 1;

#if 0
{
    tut_packhdr(send_packet, 0, 0, TUT_ERQ);
    if(sendto_in(usfd, send_packet, 32, &conn_addr) < 0)
    {
	perror("sendto");
    }
    if((recv_size = recvfrom_in(usfd, recv_packet, MAXPACKETSIZE,
				&recv_addr)) < 0);
    puts("## OK");
}
#endif

    /* TCP  3-Way Handshake Ԥ
     * Ʊˡ UDP ݡȤ³ TCP ݡȤ롣
     *  UDP ݡȤϡACK+SYN Ʊ˼롣
     *
     * SYN äơᥤ롼פ롣
     */
    tut_packhdr(send_packet, mseq++, 0, TUT_SYN);
    do id = rand32(); while(id == 0);

#if 0
    pack(send_packet + HDRSIZE, "nnNN", tut_port, port,
	 ntohl(tcp_addr), id);
#else
    pack(send_packet + HDRSIZE, "nnNN", bind_port, port,
	 ntohl(tcp_addr), id);
#endif
    send_size = HDRSIZE + 12;

    if(sendto_in(usfd, send_packet, send_size, &conn_addr) < 0)
    {
	perror("sendto");
    }
    status = TUT_SYN_SENT;
    wait_usec_time = INIT_WAIT_USEC;
    resend_cnt = 0;

    while(getppid() != 1) /* ӤˤʤäƤޤä exit  */
    {
	if(resend_cnt > MAX_RESEND)
	    break;
	fdset = orig_fdset;

#if CHECK
	/* send_size == 0 ΤȤϡǡʤΤǡɤ߹ߤǽ
	 * ʤޤǡ֥å롣
	 * send_size > 0 ΤȤϡǡΤǡwait_usec_time == 0 
	 * äƤϤʤʤ
	 */
	if(((send_size == 0) ^ (wait_usec_time == 0)) == 0 ||
	   status != TUT_ESTABLISHED)
	    /* OK */;
	else
	    fprintf(stderr, "BUG?\n");
#endif

#if CHECK
	printf("wait_usec_time=%u ", wait_usec_time);
	printf("send_size=%d ", send_size);
	printf("resend_cnt=%d ", resend_cnt);
	printf("status=%d\n", status);
#endif

	if(wait_usec_time)
	{
	    timeout.tv_sec  = wait_usec_time / 1000000;
	    timeout.tv_usec = wait_usec_time % 1000000;
	    cnt = select(nfds, &fdset, NULL, NULL, &timeout);
	}
	else
	    cnt = select(nfds, &fdset, NULL, NULL, NULL);

	if(cnt < 0)
	{
	    perror("select");
	    sleep(1);
	    continue;
	}

	if(cnt == 0 && wait_usec_time > 0)
	{
	    timeout_flag = 1;
	    recv_size = 0;
#if CHECK
	    printf("Timeout\n");
#endif
	}
	else
	{
	    timeout_flag = 0;
	    if(FD_ISSET(usfd, &fdset))
	    {
		if((recv_size = recvfrom_in(usfd, recv_packet, MAXPACKETSIZE,
					    &recv_addr)) < 0)
		{
		    perror("recvfrom");
		    close(usfd); close(wfd); close(rfd); exit(1);
		}
		if(recv_size < HDRSIZE)
		    continue;
		tut_unpackhdr(recv_packet, &recv_seq, &recv_ack, &recv_flags);
#if CHECK
		printf("recv_size=%d mseq=%d yseq=%d recv_seq=%d recv_ack=%d flags=0x%x\n",
		       recv_size, mseq, yseq, recv_seq, recv_ack, recv_flags);
#endif
		if(IS_FLAG_SET(recv_flags, TUT_ERQ))
		{
		    pack(recv_packet + 4, "N", TUT_ERY);
		    sendto_in(usfd, recv_packet, recv_size, &recv_addr);
		    continue;
		}
	    }
	    else
	    {
		recv_size = 0;
	    }
	}

#if CHECK
	printf("status=%d\n", status);
#endif
	if(status == TUT_SYN_SENT)
	{
	    if(timeout_flag)
	    {
		if(readable(rfd) && fionread(rfd) == 0)
		{
		    /* CHECK & FIXME */
		    fprintf(stderr, "Connection closed\n");
		    close(usfd); close(wfd); close(rfd); exit(1);
		}

		/* ॢ
		 * SYN ѥåȤ
		 */
		if(sendto_in(usfd, send_packet, send_size, &conn_addr) < 0)
		{
		    perror("sendto");
		}
#if CHECK
		printf("resend_cnt: %d\n", resend_cnt);
#endif
		resend_cnt++;
		continue;
	    }

	    if(recv_ack != mseq)
		continue;

	    if(IS_FLAG_SET(recv_flags, TUT_ACK | TUT_RST))
	    {
		fdprintf(wfd, "Can't connect\n");
		close(usfd); close(wfd); close(rfd); exit(1);
	    }

	    if(!IS_FLAG_SET(recv_flags, TUT_ACK | TUT_SYN))
		continue;

	    if(recv_size == HDRSIZE + 2)
	    {
		unpack(recv_packet + HDRSIZE, "n", &us);
		yipaddr = ipaddr;
	    }
	    else
	    {
		unpack(recv_packet + HDRSIZE, "nnN", &us, &usdmy, &yipaddr);
	    }
	    yipaddr = htonl(yipaddr);
#if CHECK
	    printf("your UDP port=%u, addr=0x%08x\n", us, yipaddr);
#endif
	    pack_sockaddr_in(&yaddr, us, yipaddr);

	    /* ACK ֤ơESTABLISHED ֤ */
	    yseq = recv_seq + 1;
	    tut_packhdr(recv_packet, 0, yseq, TUT_ACK);
	    if(sendto_in(usfd, recv_packet, HDRSIZE, &yaddr) < 0)
	    {
		perror("sendto");
	    }

	    /* rfd ɤߤȤ褦ˤ */
	    FD_SET(rfd, &orig_fdset);
	    if(rfd >= nfds)
		nfds = rfd + 1;

	    /* ʤ */
	    send_size = 0;
	    wait_usec_time = 0;

	    /* TUT_ESTABLISHED  */
	    status = TUT_ESTABLISHED;
	    resend_cnt = 0;

	    established(id, rfd, wfd, usfd, &yaddr, mseq, yseq);
	}
    }
#if CHECK
    printf("## Out of mainloop.\n");
#endif
    close(usfd); close(wfd); close(rfd);
    exit(1);
}

static char *parse_relay_arg(const char* arg, U16* bind_port, U16* port)
{
    static char buff[BUFSIZ];
    char *p, *q;

    *bind_port = 0;
    memcpy(buff, arg, strlen(arg) + 1);
    for(p = buff; *p && *p != ':'; p++)
	;
    if(*p != ':')
	return NULL;
    *p++ = '\0';
    *port = atoi(p);
    if((q = strchr(p, ',')) != NULL)
	*bind_port = atoi(q + 1);

    return buff;
}
