#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <errno.h>
#ifdef sun
extern int errno;
#endif /* sun */

#include "net.h"
#include "tut.h"

#ifndef CHECK
#define CHECK 0
#endif

#define MAXCONNECT 64

static int conn_usfd;
static struct sockaddr_in conn_maddr;
/*void doit(U32 id, U32 ipaddr, U16 udp_port, U16 bind_port, U16 tcp_port, U32 addr, U16 yseq);*/
void doit(U32 id, U16 uport, int usfd,
	  struct sockaddr_in* yaddr,
	  int fd, U16 yseq);
char recv_packet[MAXPACKETSIZE+SPARESIZE];
char send_packet[MAXPACKETSIZE+SPARESIZE];
U32 myipaddr;

static struct
{
    U32 id;
    int pid;
    U32 addr;
    U16 rcv_port;
    U16 session_port;
} connect_id[MAXCONNECT];

static int connect_reset(struct sockaddr_in* to, U16 seq)
{
    char buff[HDRSIZE];

    tut_packhdr(buff, 0, seq, TUT_ACK | TUT_RST);
    return sendto_in(conn_usfd, buff, HDRSIZE, to);
}

static int new_udpsock(struct sockaddr_in *addr)
{
    int usfd, n;

    if((usfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
	perror("socket");
	return -1;
    }
    so_reuseaddr(usfd);

    pack_sockaddr_in(addr, 0, htonl(INADDR_ANY));
    if(bind(usfd, (struct sockaddr *)addr, sizeof(*addr)) < 0)
    {
	perror("bind");
	close(usfd);
	return -1;
    }
    if(getsockname(usfd, (struct sockaddr *)addr, &n) < 0)
	perror("Warning: getsockname");

    return usfd;
}

static void syn_reply(U16 port, U16 seq, struct sockaddr_in* yaddr)
{
    char buff[HDRSIZE + 8];

#if CHECK
    printf("##%d SYN_ACK\n", getpid());
#endif

    port = ntohs(port);
    tut_packhdr(buff, 0, seq, TUT_ACK | TUT_SYN);
    pack(buff + HDRSIZE, "nnN", port, 0, ntohl(myipaddr));
    sendto_in(conn_usfd, buff, HDRSIZE + 8, yaddr);
}

void sigchld_handler(int sig)
{
    int status, pid, i;

    while((pid = wait3(&status, WNOHANG, NULL)) > 0)
    {
	for(i = 0; i < MAXCONNECT; i++)
	{
	    if(connect_id[i].pid == pid)
	    {
		puts("######################################");
		printf("## EXIT: free entry=%d, ID=0x%08x\n",
		       i, connect_id[i].id);
		puts("######################################");

		connect_id[i].pid = -1;
		connect_id[i].id  = 0;
		connect_id[i].addr = 0;
		connect_id[i].rcv_port = 0;
		connect_id[i].session_port = 0;
		break;
	    }
	}
    }
    signal(SIGCHLD, sigchld_handler);
}

void wait_child(int pid)
{
    int status;

    while(wait(&status) != pid)
	;
}

void usage(void)
{
    fprintf(stderr, "Usage: ut [-d] [-h] [-i <interface_address>]\n");
    exit(1);
}

U32 interface_address(const char* name)
{
    char buff[64];
    struct hostent *hp;
    U32 addr;

    if(name == NULL)
    {
	if(gethostname(buff, sizeof(buff)) < 0 || buff[0] == '\0')
	{
	    fprintf(stderr, "Can't get local host IP address.\n");
	    fprintf(stderr, "Please specify the address `-i' option.\n");
	    exit(1);
	}
	name = buff;
    }

    addr = name2addr(name);

    if(addr == INADDR_NONE)
    {
	fprintf(stderr, "Can't get IP address: <%s>\n", name);
	exit(1);
    }

    return addr;
}

void main(int argc, char** argv)
{
    int i;
    U32 tcp_addr;
    int dflag = 0;
    char* tcp_interfaceaddr = NULL;

    /* parse args */
    for(i = 1; i < argc; i++)
    {
	if(!strcmp(argv[i], "-h"))
	{
	    usage();
	}
	else if(!strcmp(argv[i], "-d"))
	{
	    dflag = 1;
	}
	else if(!strcmp(argv[i], "-i"))
	{
	    tcp_interfaceaddr = argv[++i];
	}
	else if(argv[i][0] == '-')
	{
	    usage();
	}
	else
	{
	    break;
	}
    }

    if(dflag)
	daemon_init();

    myipaddr = interface_address(tcp_interfaceaddr);

    if((conn_usfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("socket");
	exit(1);
    }
    so_reuseaddr(conn_usfd);
    pack_sockaddr_in(&conn_maddr, UU_CONN_PORT, htonl(INADDR_ANY));
    if(bind_in(conn_usfd, &conn_maddr) < 0)
    {
	perror("bind");
	close(conn_usfd);
	exit(1);
    }

    for(i = 0; i < MAXCONNECT; i++)
    {
	connect_id[i].id = 0;
	connect_id[i].pid = -1;
    }

#ifndef SINGLE_USER
    signal(SIGCHLD, sigchld_handler);
#endif /* SINGLE_USER */

    while(1)
    {
	int n, conn_entry, pid;
	struct sockaddr_in addr, maddr, taddr;
	U16 udp_port, bind_port, tcp_port, recv_seq, recv_ack;
	U32 id, ipaddr, recv_flags;
	int session_usfd;
	int tcp_fd;
	
      recv_next:
	bind_port = 0;
#if 0
	puts("## Recvfrom");
#endif
	if((n = recvfrom_in(conn_usfd, recv_packet, MAXPACKETSIZE, &addr)) < 0)
	{
	    perror("recvfrom");
	    close(conn_usfd);
	    exit(1);
	}
	tut_unpackhdr(recv_packet, &recv_seq, &recv_ack, &recv_flags);

	if(IS_FLAG_SET(recv_flags, TUT_ERQ))
	{
	    pack(recv_packet + 4, "N", TUT_ERY);
	    sendto_in(conn_usfd, recv_packet, n, &addr);
	    continue;
	}

	if(!IS_FLAG_SET(recv_flags, TUT_SYN))
	    continue;

	ipaddr = addr.sin_addr.s_addr;
	udp_port = ntohs(addr.sin_port);

	unpack(recv_packet + HDRSIZE, "nnNN", &bind_port, &tcp_port,
	       &tcp_addr, &id);
	tcp_addr = htonl(tcp_addr);

#ifdef SECURE_UT
	bind_port = 0;
#endif /* SECURE_UT */


	for(i = 0; i < MAXCONNECT; i++)
	{
#if 0
	    if(connect_id[i].id == id)
		goto recv_next;
#endif
	    if(connect_id[i].addr == ipaddr &&
	       connect_id[i].rcv_port == udp_port)
	    {
		syn_reply(connect_id[i].session_port, recv_seq + 1, &addr);
		goto recv_next;
	    }
	}

	for(conn_entry = 0; conn_entry < MAXCONNECT; conn_entry++)
	    if(connect_id[conn_entry].id == 0)
		break;
	if(conn_entry == MAXCONNECT)
	{
	    connect_reset(&addr, recv_seq + 1);
	    continue;
	}


	if((session_usfd = new_udpsock(&maddr)) < 0)
	{
	    connect_reset(&addr, recv_seq + 1);
	    continue;
	}

	/* tcp_port ³롣
	 * Ԥȡipaddr  udp_port ءRST åꡢλ롣
	 */
	if(tcp_addr == 0)
	    tcp_addr = name2addr("127.0.0.1");
#if CHECK
	printf("## Try to connect (port=%d, addr=%s)\n",
	       tcp_port, inet_ntoa(*((struct in_addr *)&tcp_addr)));
#endif

	if((tcp_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	    perror("socket");
	    connect_reset(&addr, recv_seq + 1);
	    continue;
	}

#if 0
	pack_sockaddr_in(&taddr, bind_port, htonl(INADDR_ANY));
#else
	pack_sockaddr_in(&taddr, bind_port, myipaddr);
#endif
	if(bind_in(tcp_fd, &taddr) < 0)
	{
	    perror("bind");
	    connect_reset(&addr, recv_seq + 1);
	    continue;
	}

	pack_sockaddr_in(&taddr, tcp_port, tcp_addr);
	if(connect_in(tcp_fd, &taddr) < 0)
	{
	    perror("connect");
	    connect_reset(&addr, recv_seq + 1);
	    continue;
	}
#if 0
	getsockname(tcp_fd, (struct sockaddr *)&taddr, &n);
	myipaddr = taddr.sin_addr.s_addr;
#endif

#if CHECK
	printf("## myipaddr=%x\n", myipaddr);
#endif

#if 0
	puts("######################################");
	printf("## FORK: reserve entry=%d, ID=%u\n", conn_entry, id);
	puts("######################################");
#endif

	if((pid = fork()) < 0)
	{
	    perror("fork");
	    exit(1);
	}

	if(pid == 0) /* child */
	{
	    doit(id, maddr.sin_port, session_usfd, &addr,
		 tcp_fd, recv_seq + 1);

/* id, ipaddr, udp_port, bind_port, tcp_port, tcp_addr, recv_seq + 1); */
	}

	/* parent */
	close(session_usfd);
	close(tcp_fd);
	connect_id[conn_entry].id = id;
	connect_id[conn_entry].pid = pid;
	connect_id[conn_entry].addr = ipaddr;
	connect_id[conn_entry].rcv_port = udp_port;
	connect_id[conn_entry].session_port = maddr.sin_port;
#ifdef SINGLE_USER
	wait_child(pid);
#endif /* SINGLE_USER */
    }
}


void doit(U32 id, U16 uport, int usfd,
	  struct sockaddr_in* yaddr,
	  int fd, U16 yseq)
{
    enum tut_status_t status;
    int send_size, recv_size;
    unsigned long wait_usec_time;
    struct sockaddr_in maddr;
    struct sockaddr_in recv_addr;

    fd_set orig_fdset, fdset;	/* for select */
    int nfds, cnt;
    struct timeval timeout;
    int timeout_flag;

    U16 mseq = 1;	/* my sequense number */

    U16 recv_seq;	/* received seq number */
    U16 recv_ack;	/* received ack number */
    U32 recv_flags;	/* received flags */
    int n;
    U16 tut_port;
    int resend_cnt;		/* β */


    syn_reply(uport, yseq, yaddr);


    wait_usec_time = INIT_WAIT_USEC;
    /* fdset ν */
    FD_ZERO(&orig_fdset);
    FD_SET(usfd, &orig_fdset);
    nfds = usfd + 1;

    status = TUT_SYN_RECEIVED;

    resend_cnt = 0;
    while(getppid() != 1)
    {
	if(resend_cnt > MAX_RESEND)
	    break;
	fdset = orig_fdset;

	if(wait_usec_time)
	{
	    timeout.tv_sec  = wait_usec_time / 1000000;
	    timeout.tv_usec = wait_usec_time % 1000000;
	    cnt = select(nfds, &fdset, NULL, NULL, &timeout);
	}
	else
	    cnt = select(nfds, &fdset, NULL, NULL, NULL);

	if(cnt < 0)
	{
	    perror("select");
	    sleep(1);
	    continue;
	}

	if(cnt == 0 && wait_usec_time > 0)
	{
	    timeout_flag = 1;
	    recv_size = 0;
	}
	else
	{
	    timeout_flag = 0;
	    if(FD_ISSET(usfd, &fdset))
	    {
		if((recv_size = recvfrom_in(usfd, recv_packet, MAXPACKETSIZE,
					    &recv_addr)) < 0)
		{
		    perror("recvfrom");
		    close(usfd); close(fd); exit(0);
		}
		if(recv_size < HDRSIZE)
		    continue;
		tut_unpackhdr(recv_packet, &recv_seq, &recv_ack, &recv_flags);
#if CHECK
		printf("recv_size=%d mseq=%d, yseq=%d, recv_seq=%d, recv_ack=%d, flags=0x%x\n",
		       recv_size, mseq, yseq, recv_seq, recv_ack, recv_flags);
#endif
		if(IS_FLAG_SET(recv_flags, TUT_ERQ))
		{
		    pack(recv_packet + 4, "N", TUT_ERY);
		    sendto_in(usfd, recv_packet, recv_size, &recv_addr);
		    continue;
		}
	    }
	    else
	    {
		recv_size = 0;
	    }
	}

	if(status == TUT_SYN_RECEIVED)
	{
	    if(timeout_flag)
	    {
		syn_reply(uport, yseq, yaddr);
#if 0
		sendto_in(usfd, send_packet, send_size, &yaddr);
#endif
		resend_cnt++;
		continue;
	    }

	    if(!IS_FLAG_SET(recv_flags, TUT_ACK) || mseq != recv_ack)
		continue;

	    /* fd ɤߤȤ褦ˤ */
	    FD_SET(fd, &orig_fdset);
	    if(fd >= nfds)
		nfds = fd + 1;

	    /* ʤ */
	    send_size = 0;
	    wait_usec_time = 0;

	    /* TUT_ESTABLISHED  */
	    status = TUT_ESTABLISHED;
	    *yaddr = recv_addr;
	    established(id, fd, fd, usfd, yaddr, mseq, yseq);
	}
    }
    exit(0);
}
