#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <stdarg.h>

#include "tut.h"

#define VA_START va_start

/* #define RAND_SEED 858662223 */

int fdgetc(int fd)
{
    unsigned char c;

    if(read(fd, &c, 1) <= 0)
	return -1;
    return (int)c;
}

void tut_packhdr(char* packet,
		 U16 seqno, U16 ackno,
		 U32 flags)
{
    ((U16 *)packet)[0] = htons(seqno);
    ((U16 *)packet)[1] = htons(ackno);
    ((U32 *)packet)[1] = htonl(flags);

#if 0
    printf("## tut_pack: %u %u %u => ", seqno, ackno, flags); xdump(packet, 8); putchar('\n');
#endif
}

void tut_unpackhdr(const char *packet,
		   U16 *seqno, U16 *ackno,
		   U32 *flags)
{
    *seqno = ((U16 *)packet)[0];
    *seqno = ntohs(*seqno);
    *ackno = ((U16 *)packet)[1];
    *ackno = ntohs(*ackno);
    *flags = ((U32 *)packet)[1];
    *flags = ntohl(*flags);

#if 0
    printf("## tut_unpack: "); xdump(packet, 8); printf(" => %u %u %u\n", *seqno, *ackno, *flags);
#endif
}

void fdprintf(int fd, const char* fmt, VADCL)
{
    va_list args;
    char buff[BUFSIZ];
    VA_START(args, fmt);
    vsprintf(buff, fmt, args);
    va_end(args);
    write(fd, buff, strlen(buff));
}

#ifndef HAS_STRERROR
extern char *sys_errlist[];
#define strerror(no) (sys_errlist[no]);
#endif /* HAS_STRERROR */

void fdperror(int fd, const char* s)
{
    struct iovec *v;
    struct iovec iov[4];

    v = iov;
    if (s && *s) {
	v->iov_base = (char *)s;
	v->iov_len = strlen(s);
	v++;
	v->iov_base = ": ";
	v->iov_len = 2;
	v++;
    }
    v->iov_base = strerror(errno); /* (caddr_t)strerror(errno); */
    v->iov_len = strlen(v->iov_base);
    v++;
    v->iov_base = "\n";
    v->iov_len = 1;
    (void)writev(fd, iov, (v - iov) + 1);
}

int wait_usec(unsigned long usec)
{
    return tut_select(0, NULL, usec);
}

int fionread(int fd)
{
    int read_size;
    if(ioctl(fd, FIONREAD, &read_size) < 0)
	return -1;
    return read_size;
}

U32 rand8(void)
{
    static int init = 0;

    if(!init) {
#ifdef RAND_SEED
	srand(RAND_SEED);
#else
	time_t t;
	t = time(&t);
	srand(t);
#endif /* RAND_SEED */
	init = 1;
    }

    /* ̥ӥåȤȤƤɤʤ */
    return (U32)((rand() >> 7) & 0xff);
}

U32 rand32(void)
{
    return rand8() | rand8() << 8 | rand8() << 16 | rand8() << 24;
}

int readable(int fd)
{
    struct timeval tv;
    fd_set fds;

    tv.tv_sec = tv.tv_usec = 0;
    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    return select(fd + 1, &fds, NULL, NULL, &tv);
}

int pack(char* buff, const char* fmt, VADCL)
{
    va_list argp;
    unsigned char c;
    U16 s;
    U32 l;
    char* p = buff;
    const char* f = fmt;

    VA_START(argp, fmt);
    while(*f)
    {
	switch(*f)
	{
	  case 'c':
	    *p++ = va_arg(argp, unsigned char);
	    break;
	  case 'n':

#if 0
#ifdef sun /* Why ? */
	    s = va_arg(argp, int);
#else
	    s = va_arg(argp, U16);
#endif /* sun */
#endif

	    s = va_arg(argp, int);

	    *p++ = (s >> 8);
	    *p++ = (s & 0xff);

	    break;
	  case 'N':
	    l = va_arg(argp, U32);
	    *p++ = (l >> 24);
	    *p++ = ((l >> 16) & 0xff);
	    *p++ = ((l >> 8) & 0xff);
	    *p++ = (l & 0xff);

	    break;
	}
	f++;
    }

    va_end(argp);

    return (char *)p - buff;
}

int unpack(const char* data, const char* fmt, VADCL)
{
    va_list argp;
    unsigned char* c;
    U16* s;
    U32* l;
    unsigned char* p = (unsigned char *)data;

    VA_START(argp, fmt);
    while(*fmt)
    {
	switch(*fmt)
	{
	  case 'c':
	    c = va_arg(argp, unsigned char *);
	    if(c != NULL)
		*c = *p;
	    p++;
	    break;
	  case 'n':
	    s = va_arg(argp, U16 *);
	    if(s != NULL)
		*s = ((U16)p[1] | (U16)p[0] << 8);
	    p += sizeof(U16);
	    break;
	  case 'N':
	    l = va_arg(argp, U32 *);
	    if(l != NULL)
		*l = (((U32)p[3]      ) | ((U32)p[2] <<  8) |
		      ((U32)p[1] << 16) | ((U32)p[0] << 24));
	    p += sizeof(U32);
	    break;
	}
	fmt++;
    }

    va_end(argp);

    return (char *)p - data;
}

void xdump(const void* val, int size)
{
    int i;
    const unsigned char* p = (const unsigned char *)val;
    unsigned char c;

    for(i = 0; i < size; i++)
    {
	c = p[i];
#if 1
/*	c = c >> 4 | c << 4; */
	if(i > 0 && i % 4 == 0)
	    putchar(' ');
	printf("%02x", c);
#else
	if(c >= ' ' && c < 128)
	    putchar(c);
	else
	    putchar('.');
#endif
    }
}

int tut_select(int maxfd, fd_set* rfds, unsigned long wait_usec_time)
{
    int cnt;
    struct timeval timeout;

    timeout.tv_sec  = wait_usec_time / 1000000;
    timeout.tv_usec = wait_usec_time % 1000000;
    do
	cnt = select(maxfd, rfds, NULL, NULL, &timeout);
    while(cnt < 0 && errno == EINTR);
    return cnt;
}

int va_init_fdset(fd_set* fds, VADCL)
{
    va_list argp;
    int maxfd, fd;

    maxfd = 0;
    FD_ZERO(fds);
    VA_START(argp, fds);
    while((fd = va_arg(argp, int)) >= 0)
    {
	if(fd > maxfd)
	    maxfd = fd;
	FD_SET(fd, fds);
    }
    return maxfd + 1;
}

void daemon_init(void)
{
    if(fork())
	exit(0);
    setsid();
}

#if 0
void sigint(void)
{
    write(1, "SIGINT\n", 7);
}

void main(void)
{
    fd_set rfds;
    int cnt;
    char buff[BUFSIZ];

    signal(SIGINT, sigint);
    while(1)
    {
	FD_ZERO(&rfds);
	FD_SET(0, &rfds);
	cnt = tut_select(1, &rfds, 3000000);
	if(cnt > 0)
	    read(0, buff, BUFSIZ);
	else
	    buff[0] = '\0';
	printf("## cnt=%d, <%s>\n", cnt, buff);
    }
}
#endif
